import numpy as np
import tensorflow as tf
from matplotlib import pyplot as plt
import pandas as pd
import os
import sys
sys.path.append(os.getcwd())
from TextClassification.argsparse_classification_preamble import argsparse_preamble
import TextClassification.classification_metrics as cls_metrics
import time
from database_preparation.utils_labeled_datasets import get_splits_for_cross_val
from database_preparation.utils_labeled_datasets import text_label_files_to_labeled_dataset
from database_preparation.preprocess import print_meta_data


def save_graphs(history, metric, save_path):
    plt.close()
    plt.plot(history.history[metric])
    plt.plot(history.history['val_' + metric])
    plt.xlabel("Epochs")
    plt.ylabel(metric)
    plt.legend([metric, 'val_' + metric])
    #plt.show()
    plt.savefig(save_path, dpi=300)
    print("generated "+save_path)


def dict2tf_dataset(dict):
    return tf.data.TextLineDataset.from_tensor_slices(([text.lower() for text in dict['text']], dict['label']))


def train_test_updatemetrics(train_dataset, test_dataset, num_classes, metrics,
                             epochs=10, plot_loss=False,
                             plot_save_path="TextClassification/plots/RNN/RNN_loss.png"):
    y_test = np.asarray(test_dataset['label'])

    train_dataset = dict2tf_dataset(train_dataset)
    test_dataset = dict2tf_dataset(test_dataset)

    # Next shuffle the data for training and create batches of these (text, label) pairs:
    BUFFER_SIZE = 10000
    BATCH_SIZE = 64
    train_dataset = train_dataset.shuffle(BUFFER_SIZE).batch(BATCH_SIZE).prefetch(tf.data.AUTOTUNE)
    test_dataset = test_dataset.batch(BATCH_SIZE).prefetch(tf.data.AUTOTUNE)

    for example, label in train_dataset.take(1):
        '''print('text: ', example.numpy())
        print('label: ', label.numpy())'''
        pass

    ############### Create the text encoder ###################
    VOCAB_SIZE = 5000
    encoder = tf.keras.layers.TextVectorization(
        max_tokens=VOCAB_SIZE)
    encoder.adapt(train_dataset.map(lambda text, label: text))

    vocab = np.array(encoder.get_vocabulary())
    '''print("vocab info:")
    print(vocab[:20])
    print(len(encoder.get_vocabulary()))

    encoded_example = encoder(example)[:3].numpy()
    print(encoded_example)

    for n in range(3):
        print("Original: ", example[n].numpy())
        print("Round-trip: ", " ".join(vocab[encoded_example[n]]))
        print()'''

    ##### create the model: #####
    embedding_dim = 64
    model = tf.keras.Sequential([
        encoder,
        tf.keras.layers.Embedding(
            input_dim=len(encoder.get_vocabulary()),
            output_dim=embedding_dim,
            # Use masking to handle the variable sequence lengths
            mask_zero=True),
        tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(embedding_dim)),
        tf.keras.layers.Dense(embedding_dim, activation='relu'),
        tf.keras.layers.Dense(num_classes, activation='softmax')
    ])
    # model.summary()

    # stacking 2 LSTM layers (seems to be much worse):
    '''model = tf.keras.Sequential([
        encoder,
        tf.keras.layers.Embedding(len(encoder.get_vocabulary()), 64, mask_zero=True),
        tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(64,  return_sequences=True)),
        tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(32)),
        tf.keras.layers.Dense(64, activation='relu'),
        tf.keras.layers.Dropout(0.5),
        tf.keras.layers.Dense(num_classes, activation='softmax')
    ])'''

    # All the layers after the Embedding support masking:
    # print([layer.supports_masking for layer in model.layers])

    # predict on a sample text without padding.
    '''print("test prediction:")
    sample_text = ('The movie was cool. The animation and the graphics '
                   'were out of this world. I would recommend this movie.')
    predictions = model.predict(np.array([sample_text]))
    print(predictions)'''

    model.compile(loss='sparse_categorical_crossentropy',
                  optimizer=tf.keras.optimizers.Adam(1e-4),
                  metrics=["accuracy"])

    start = time.time()

    # evaluate:
    if plot_loss:
        history = model.fit(train_dataset, epochs=epochs,
                            validation_data=test_dataset,
                            validation_steps=2)
        test_loss, test_acc = model.evaluate(test_dataset)
        print('Test Loss:', test_loss)
        print('Test Accuracy:', test_acc)
        #save_graphs(history, 'accuracy')
        save_graphs(history, 'loss', plot_save_path)

    else:
        model.fit(train_dataset, epochs=epochs)
        predictions = model.predict(test_dataset)
        y_pred = np.argmax(predictions, axis=-1)
        metrics.update_metrics(y_test, y_pred, True, start)


def main():
    ############# get labeled text data ###################

    # how to convert words 2 ids with gensim:
    # words = corpora.Dictionary(diag_lst)

    args = argsparse_preamble()

    print("RNN Evaluation with corpus " + args.path2corpus + " and cluster set " + args.clustered_data)
    print("infos about corpus:")
    print_meta_data(args.path2corpus.replace('.pkl', '_meta.pkl'))

    # dataset = load_labeled_dataset(args.clustered_data)
    dataset = text_label_files_to_labeled_dataset(args.clustered_data, path2corpus=args.path2corpus)

    num_classes = int(pd.DataFrame(dataset["label"]).nunique())

    metrics = cls_metrics.ClassificationMetrics("RNN")

    folds = 10
    epochs = 70

    for i, (train_dataset, test_dataset) in enumerate(get_splits_for_cross_val(dataset, folds)):
        if args.loss_curve_check:
            train_test_updatemetrics(train_dataset, test_dataset, num_classes,
                                     metrics, epochs=epochs, plot_loss=True,
                                     plot_save_path="TextClassification/plots/RNN/RNN_loss_"+str(i+1)+".png")
        else:
            print("====== RNN train/test run " + str(i + 1) + "/" + str(folds) + " ======")
            print(str(len(train_dataset["label"])) + " train documents")
            print(str(len(test_dataset["label"])) + " test documents")
            train_test_updatemetrics(train_dataset, test_dataset, num_classes, metrics, epochs=epochs)

    if not args.loss_curve_check:
        metrics.save_scores_to_disk(args.clustered_data)
        metrics.pickle_object(args.clustered_data)
        cls_metrics.print_results_as_latextable(metrics.json_file_path)


if __name__ == '__main__':
    main()
