
'''
this script splits the reports (.txt files) located in the
--path_to_reports folder into description-, diagnosis- and end-section
and saves these parts in --target_folder_path.

In addition, a pandas dataframe (--df_cases_file) is then generated in
which it is saved which 3 report-sections belong together. The df_cases
dataframe is also there to label the reports with different labelsets.
Also pass --author_names (space separated names, cased) to label the reports by found
authors (stored in df_cases)

The passed reports can also be labeled.
The label of each report is expected to be in the title/name of the .txt file,
located after the start symbol "--label_in_filename_symbol" until .txt
'''

import sys
import glob
from tqdm import tqdm
import pandas as pd
from database_preparation.reportPreparationTools import get3parts
import os, shutil
import argparse
from database_preparation.label_reports_with_authors import add_author_labels_to_df_cases
sys.path.append(os.getcwd())

# parse arguments:
sys.path.append(os.getcwd())
parser = argparse.ArgumentParser()
parser.add_argument("--path_to_reports",
                    default='../DataNephroTexts/reports')
parser.add_argument("--target_folder_path",
                    default='../DataNephroTexts')
parser.add_argument("--df_cases_file",
                    default='database/df_cases.pkl')
parser.add_argument("--author_names",
                        default="Name1 Name2")
parser.add_argument("--text_encoding",
                        default="latin-1")
parser.add_argument("--use_newest_reports", action='store_true')
parser.add_argument("--label_in_filename_symbol",
                        default='')
parser.add_argument("--save_labels_as_ints", action='store_true')
args = parser.parse_args()

# %% get all files
# get the primary reports, which histonums are ending wit .0:
report_file_list = glob.glob(args.path_to_reports + '/*0.txt')

if len(report_file_list) <= 0:
    report_file_list = glob.glob(args.path_to_reports + '/*.txt')
    if len(report_file_list) == 0:
        print(f"ERROR: Cant find .txt files in {args.path_to_reports}.")
        exit(1)
    print(f'Using all reports, located in {args.path_to_reports}')

else:
    # get all last reports (newest ones)
    if args.use_newest_reports:
        print("using newest reports (if there is a H.123.0 and a H.123.2, take H.123.2)")
        for idx in tqdm(range(0, len(report_file_list))):
            t_file_name = report_file_list[idx]
            report_file_list[idx] = glob.glob(t_file_name[0:-5] + "*.txt")[-1]
    else:
        print("using oldes reports (if there is a H.123.0 and a H.123.2, take H.123.0)")



def make_folder(dir):
    shutil.rmtree(dir, ignore_errors=True)
    os.makedirs(dir)


make_folder(args.target_folder_path + "/description")
make_folder(args.target_folder_path + "/diagnosis")
make_folder(args.target_folder_path + "/end")
make_folder(args.target_folder_path + "/label")

print(f"Splitting reports of corpus {args.path_to_reports} into description-, diagnosis- and end-sections...")

# %% iterate over all files
error_file_list = []
no_error_file_list = []
lst_description, lst_diagnose, lst_end, lst_labels = [], [], [], []
for idx, t_file in tqdm(enumerate(report_file_list)):

    try:
        txt_micro, txt_conclusion, end = get3parts(t_file)
        no_error_file_list.append(t_file)

    except:
        #print(f"skipped file {t_file}. Could not split the text into description and diagnosis-part")
        error_file_list.append(t_file)
        continue

    with open(args.target_folder_path + "/description/description#" + str(idx) + ".txt", "w",
              encoding=args.text_encoding) as text_file:
        text_file.write(txt_micro)

    with open(args.target_folder_path + "/diagnosis/diagnosis#" + str(idx) + ".txt", "w",
              encoding=args.text_encoding) as text_file:
        text_file.write(txt_conclusion)

    with open(args.target_folder_path + "/end/end#" + str(idx) + ".txt", "w",
              encoding=args.text_encoding) as text_file:
        if end == None:
            end = "None"
        text_file.write(end)

    if args.label_in_filename_symbol != '':
        try:
            a = t_file.index('#') + 1
            b = t_file.index('.txt')
            label = t_file[a:b]
        except:
            label = 'None'

        with open(args.target_folder_path + "/label/label#" + str(idx) + ".txt", "w",
                  encoding=args.text_encoding) as text_file:
            text_file.write(label)

        # lst_labels.append('label#' + str(idx) + ".txt")
        lst_labels.append(label)

    lst_description.append('description#' + str(idx) + ".txt")
    lst_diagnose.append('diagnosis#' + str(idx) + ".txt")
    lst_end.append('end#' + str(idx) + ".txt")


#save skipped reports:
with open(args.target_folder_path + "/failed_to_split_list.txt", "w") as text_file:
    text_file.write('\n'.join(error_file_list))

# print infos:
print(" ===> finished < === ")
print(f"skipped {len(error_file_list)} reports "
      f"({round(len(error_file_list)/len(report_file_list)*100,1)}%), since text splitting "
      f"failed (see {args.target_folder_path + '/failed_to_split_list.txt'}).")
processed_docs = len(lst_description)

#print(f"saved {processed_docs} description sections at {args.target_folder_path + '/description'}")
#print(f"saved {processed_docs} diagnosis sections at {args.target_folder_path + '/diagnosis'}")
#print(f"saved {processed_docs} end sections at {args.target_folder_path + '/end'}")

# create and save df_cases file, which is also used to save the labels (=cluster-indices) for each text.
if args.label_in_filename_symbol == '':
    df = pd.DataFrame(list(zip(lst_description, lst_diagnose, lst_end)),
                  columns=['description_text_files', 'diagnosis_text_files', 'end_text_files'])
else:
    if args.save_labels_as_ints:
        class2int = dict()
        for i, c in enumerate(list(set(lst_labels))):
            class2int[c] = i
        df = pd.DataFrame(list(zip(lst_description, lst_diagnose, lst_end, [class2int[l] for l in lst_labels])),
                          columns=['description_text_files', 'diagnosis_text_files', 'end_text_files', 'label_golden'])
        with open(args.target_folder_path + "/classname2integer.txt", "w") as text_file:
            text_file.write('\n'.join([f'{class2int[name]}\t\t{name}' for name in list(set(lst_labels))]))
    else:
        df = pd.DataFrame(list(zip(lst_description, lst_diagnose, lst_end, lst_labels)),
                          columns=['description_text_files', 'diagnosis_text_files', 'end_text_files', 'label_golden'])
df.to_pickle(args.df_cases_file)

# search for authors in end-sections in order to add them as labels to the df_cases file:
if args.label_in_filename_symbol == '':
    try:
        add_author_labels_to_df_cases(args.target_folder_path + '/end', args.author_names.split(' '), args.df_cases_file)
    except:
        print("label the reports with authors failed.")

df = pd.read_pickle(args.df_cases_file)
print(f"saved df_cases at {args.df_cases_file}\n")
print(f'it looks like this:')
print(df)
print(f"first element of {args.df_cases_file}: \n{df.iloc[0]}")
print()
